# -*- coding: utf-8 -*-
"""
System specific filesystem utilities
"""
from abc import ABC
import glob
import os
import s3fs
import shutil


class FileSystemError(Exception):
    """
    Custom filesystem error
    """


class BaseFileSystem(ABC):
    """
    Abstract Base class for handling filesystem operations
    """
    def __init__(self, path):
        """
        Parameters
        ----------
        path : str
            File path
        """
        self._path = path
        self._operations = {'cp': None,
                            'exists': None,
                            'isfile': None,
                            'isdir': None,
                            'glob': None,
                            'ls': None,
                            'mkdirs': None,
                            'mv': None,
                            'open': None,
                            'rm': None,
                            'size': None,
                            'walk': None}

    def __repr__(self):
        msg = ("{} filesystem operations on {}"
               .format(self.__class__.__name__, self.path))

        return msg

    def __contains__(self, key):
        return key in self.operations

    def __getitem__(self, operation):
        """
        Get filesystem specific operation function/method

        Parameters
        ----------
        operation : str
            Filesystem operation name

        Returns
        -------
        obj
        """
        if operation not in self:
            msg = ('{} is not a valid {} filesystem operation! Please select '
                   'one of:\n{}'
                   .format(operation, self.__class__.__name__,
                           self.operations))
            raise FileSystemError(msg)

        return self._operations[operation]

    @property
    def path(self):
        """
        File path to perform filesystem operation on

        Returns
        -------
        str
        """
        return self._path

    @property
    def operations(self):
        """
        Available filesystem operations

        Returns
        -------
        list
        """
        return sorted(self._operations)


class Local(BaseFileSystem):
    """
    Local filesystem utilities
    """
    def __init__(self, path):
        """
        Parameters
        ----------
        path : str
            File or directory path
        """
        self._path = path
        self._operations = {'cp': self.cp,
                            'exists': os.path.exists,
                            'isfile': os.path.isfile,
                            'isdir': os.path.isdir,
                            'glob': glob.glob,
                            'ls': os.listdir,
                            'mkdirs': os.makedirs,
                            'mv': shutil.move,
                            'open': self._open,
                            'rm': self.rm,
                            'size': os.path.getsize,
                            'walk': os.walk}

    @staticmethod
    def cp(src, dst, **kwargs):
        """
        Copy file or recursively copy directory at src to dst

        Parameters
        ----------
        src : path
            File or directory to copy
        dst : path
            Destination to copy file or directory too
        """
        if os.path.isfile(src):
            shutil.copy(src, dst, **kwargs)
        else:
            shutil.copytree(src, dst, **kwargs)

    # pylint: disable=unused-argument
    @staticmethod
    def _open(path, mode='r', **kwargs):
        """
        Delete file or files in given directory

        Parameters
        ----------
        path : str
            File or directory path
        mode : str, optional
            mode with which to open file, by default 'r'
        kwargs : dict
            kwargs for shutil.rmtree

        Returns
        -------
        str
        """
        return path

    @staticmethod
    def rm(path, **kwargs):
        """
        Delete file or files in given directory

        Parameters
        ----------

        kwargs : dict
            kwargs for shutil.rmtree

        Returns
        -------
        str
        """
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path, **kwargs)


class S3(BaseFileSystem):
    """
    S3 filesystem utilities
    """
    def __init__(self, s3_path, anon=False, profile=None, **kwargs):
        """
        Parameters
        ----------
        s3_path : str
            S3 object path
        anon : bool, optional
            Whether to use anonymous credentials, by default False
        profile : str, optional
            AWS credentials profile, by default None
        """
        self._path = s3_path
        self._s3fs = s3fs.S3FileSystem(anon=anon, profile=profile,
                                       **kwargs)

        self._operations = {'cp': self.copy,
                            'exists': self._s3fs.exists,
                            'isfile': self._s3fs.isfile,
                            'isdir': self._s3fs.isdir,
                            'glob': self._s3fs.glob,
                            'ls': self.ls,
                            'mkdirs': self._s3fs.mkdirs,
                            'mv': self._s3fs.mv,
                            'open': self._s3fs.open,
                            'rm': self._s3fs.rm,
                            'size': self._s3fs.size,
                            'walk': self.walk}

    def copy(self, src, dst, **kwargs):
        """
        Copy file within, to and from s3

        Parameters
        ----------
        src : str
            Source path
        dst : str
            Destination path
        """
        s3_src = src.lower().startswith('s3:')
        s3_dst = dst.lower().startswith('s3:')

        if s3_src and s3_dst:
            self._s3fs.copy(src, dst, **kwargs)
        elif s3_src:
            self._s3fs.get(src, dst, **kwargs)
        elif s3_dst:
            self._s3fs.put(src, dst, **kwargs)
        else:
            msg = ("Source ({}) or destination ({}) path must be on s3!"
                   .format(src, dst))
            raise ValueError(msg)

    def ls(self, path):
        """
        List objects under path

        Parameters
        ----------
        path : str
            path to list objects under

        Returns
        -------
        list
            objects that exist under path
        """
        return [os.path.basename(obj) for obj in self._s3fs.ls(path)]

    def walk(self, path):
        """
        Walk down directory structure

        Parameters
        ----------
        path : str
            Path to walk down

        Returns
        -------
        list
            root, directories, files below path
        """
        return [('s3://' + r, d, f) for r, d, f in self._s3fs.walk(path)]



# -*- coding: utf-8 -*-
"""
Utilities to abstractly handle filesystem operations
"""
import inspect
from .filesystems import Local, S3


class FileSystem:
    """
    Class to abstract file location and allow file-system commands that are
    """
    def __init__(self, path, anon=False, profile=None, **kwargs):
        """
        Parameters
        ----------
        path : str
            S3 object path or file path
        anon : bool, optional
            Whether to use anonymous credentials, by default False
        profile : str, optional
            AWS credentials profile, by default None
        """
        self._path = path
        self._handler = None
        if path.lower().startswith('s3:'):
            self._fs = S3(path, anon=anon, profile=profile, **kwargs)
        else:
            self._fs = Local(path)

        self._check_operations()

    def __repr__(self):
        msg = ("{} operations on {}"
               .format(self.__class__.__name__, self.path))

        return msg

    def __enter__(self):
        self._handler = self.open()

        return self._handler

    def __exit__(self, type, value, traceback):
        try:
            self._handler.close()
        except AttributeError:
            pass
        except Exception:
            raise

        if type is not None:
            raise

    @property
    def path(self):
        """
        File path to perform filesystem operation on

        Returns
        -------
        str
        """
        return self._path

    def _check_operations(self):
        """
        Check to ensure the File System class being used has all of the
        required file system operations defined.
        """
        operations = [attr for attr, attr_obj
                      in inspect.getmembers(self.__class__)
                      if not attr.startswith('_')
                      and not isinstance(attr_obj, property)
                      and not inspect.ismethod(attr_obj)]

        missing = list(set(operations) - set(self._fs.operations))
        if missing:
            msg = ("The following filesystem operations are not defined in "
                   "{}:\n{}".format(missing, self._fs))
            raise NotImplementedError(msg)

        return operations

    def cp(self, dst, **kwargs):
        """
        Copy file to given destination

        Parameters
        ----------
        dst : str
            Destination path
        kwargs : dict
            kwargs for s3fs.S3FileSystem.copy
        """
        self._fs['cp'](self.path, dst, **kwargs)

    def exists(self):
        """
        Check if file path exists

        Returns
        -------
        bool
        """
        return self._fs['exists'](self.path)

    def isfile(self):
        """
        Check if path is a file

        Returns
        -------
        bool
        """
        return self._fs['isfile'](self.path)

    def isdir(self):
        """
        Check if path is a directory

        Returns
        -------
        bool
        """
        return self._fs['isdir'](self.path)

    def glob(self, **kwargs):
        """
        Find all file paths matching the given pattern

        Parameters
        ----------
        kwargs : dict
            kwargs for s3fs.S3FileSystem.glob

        Returns
        -------
        list
        """
        return self._fs['glob'](self.path, **kwargs)

    def ls(self):
        """
        List everyting under given path

        Returns
        -------
        list
        """
        return sorted(self._fs['ls'](self.path))

    def mkdirs(self, **kwargs):
        """
        Make desired directory and any intermediate directories

        Parameters
        ----------
        kwargs : dict
            kwargs for s3fs.S3FileSystem.mkdirs
        """
        self._fs['mkdirs'](self.path, **kwargs)

    def mv(self, dst, **kwargs):
        """
        Move file or all files in directory to given destination

        Parameters
        ----------
        dst : str
            Destination path
        kwargs : dict
            kwargs for s3fs.S3FileSystem.mv
        """
        self._fs['mv'](self.path, dst, **kwargs)

    def open(self, mode='rb', **kwargs):
        """
        Open S3 object and return a file-like object

        Parameters
        ----------
        mode : str
            Mode with which to open the s3 object
        kwargs : dict
            kwargs for s3fs.S3FileSystem.open

        Returns
        -------
        Return a file-like object from the filesystem
        """
        return self._fs['open'](self.path, mode=mode, **kwargs)

    def rm(self, **kwargs):
        """
        Delete file or files in given directory

        Parameters
        ----------
        kwargs : dict
            kwargs for s3fs.S3FileSystem.rm
        """
        self._fs['rm'](self.path, **kwargs)

    def size(self):
        """
        Get file size in bytes

        Returns
        -------
        float
        """
        return self._fs['size'](self.path)

    def walk(self):
        """
        Recursively search directory and all sub-directories

        Returns
        -------
        path : str
            Root path
        directory : list
            All directories in path
        file : list
            All files in path
        """
        return self._fs['walk'](self.path)

    @classmethod
    def copy(cls, src_path, dst_path, anon=False, profile=None, **kwargs):
        """
        Copy file(s) from src_path to dst_path. Either can be local or in the
        cloud.

        Parameters
        ----------
        src_path : str
            Source path to copy file(s) from, can be local or in the cloud
        dst_path : str
            Destination path to copy file(s) to, can be local or in the cloud
        anon : bool, optional
            Whether to use anonymous credentials, by default False
        profile : str, optional
            AWS credentials profile, by default None
        """
        s3 = (src_path.lower().startswith('s3:')
              or dst_path.lower().startswith('s3:'))
        if s3:
            path = 's3:'
        else:
            path = ''

        fs = cls(path, anon=anon, profile=profile, **kwargs)

        fs._fs['cp'](src_path, dst_path)



# -*- coding: utf-8 -*-
"""
setup.py
"""
import os
from codecs import open
from setuptools import setup, find_packages
from setuptools.command.develop import develop
from subprocess import check_call
import shlex
from warnings import warn

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "cloud_fs", "version.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


class PostDevelopCommand(develop):
    """
    Class to run post setup commands
    """

    def run(self):
        """
        Run method that tries to install pre-commit hooks
        """
        try:
            check_call(shlex.split("pre-commit install"))
        except Exception as e:
            warn("Unable to run 'pre-commit install': {}"
                 .format(e))

        develop.run(self)


with open("requirements.txt") as f:
    install_requires = f.readlines()

test_requires = ["pytest>=5.2", "h5py>=2.10.0,!=3.0.0"]
description = ("`cloud-fs` is a generalized file-system handler that will "
               "dynamically determine if files are local or on the cloud "
               "(currently AWS) and perform basic file-systm operations.")

setup(
    name="NREL-cloud_fs",
    version=version,
    description=description,
    long_description=readme,
    author="Michael Rossol",
    author_email="michael.rossol@nrel.gov",
    url="https://nrel.github.io/cloud_fs/",
    packages=find_packages(),
    package_dir={"cloud_fs": "cloud_fs"},
    include_package_data=True,
    license="BSD 3-Clause",
    zip_safe=False,
    keywords="cloud_fs",
    python_requires='>=3.7',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    install_requires=install_requires,
    extras_require={
        "test": test_requires,
        "dev": test_requires + ["flake8", "pre-commit", "pylint"],
    },
    cmdclass={"develop": PostDevelopCommand},
)
